import java.net.URL;
import java.awt.Image;
import javax.imageio.ImageIO;
import java.util.ArrayList;

public class Animation
{
    // note: unit of time is milliSeconds
    private int totalFrameCount;
    private int currentFrameIndex;
    private boolean recentFrameChange;
    private int internalTimer;
    private int totalDuration; // seconds
    private boolean animationLoop;
    private boolean animationFinished;
    private ArrayList<Image> frameImages;
    private ArrayList<Integer> frameEndTimes;
    
    /** Initialize animation. */
    public Animation()
    {
        totalFrameCount = 0;
        currentFrameIndex = 0;
        recentFrameChange = true;
        internalTimer = 0;
        totalDuration = 0;
        animationLoop = true;
        animationFinished  = false;
        frameImages = new ArrayList<Image>();
        frameEndTimes = new ArrayList<Integer>();
    }
    
    /** Initialize a new animation that clones an existing one. */
    /**  
     *   This function enables us to make copies of this object; needed because
     *   by default "=" assigns a reference to, not the value of, an object.
     */
    public Animation(Animation a)
    {
        if (a == null) return;
        
        totalFrameCount    = a.totalFrameCount;
        currentFrameIndex  = a.currentFrameIndex;
        recentFrameChange  = a.recentFrameChange;
        internalTimer      = a.internalTimer;
        totalDuration      = a.totalDuration;
        animationLoop      = a.animationLoop;
        animationFinished  = a.animationFinished;
        frameImages        = a.frameImages;
        frameEndTimes      = a.frameEndTimes; 
    }
    
    public void addFrame(Image newImage, int imageDuration)
    {
        totalFrameCount++;
        totalDuration += imageDuration;
        frameImages.add(newImage);
        frameEndTimes.add(totalDuration);
        //System.out.println("Frame added, duration " + imageDuration
    }
    
    public void addFrame(String imageFileName, int imageDuration)
    {   addFrame( loadImage(imageFileName), imageDuration );   }
    
    public void setAnimationLoop(boolean b)
    {   animationLoop = b;   }
    
    // if animation loops, it will never be finished.
    public boolean isAnimationFinished()
    {   return (!animationLoop && animationFinished);   }
    
    /** Did the animation frame change on the most recent update? */
    public boolean frameChanged()
    {   return recentFrameChange;   }
    
    /** Change animation frame, if necessary. */
    public void update(int elapsedTime) // time in milliSeconds
    {   
        if ( (totalFrameCount <= 1) || isAnimationFinished() )
            return; // no images will be changing
            
        recentFrameChange = false;
        // update internal timer
        internalTimer += elapsedTime;
        
        // restart animation if necessary
        if (internalTimer >= totalDuration)  
        {
            if (animationLoop)
            {
                internalTimer = internalTimer % totalDuration;
                currentFrameIndex = 0;
                recentFrameChange = true;
            }
            else // reached the end of loop
            {
                animationFinished = true;
                return; // this will be the final update
            }
        }

        // update current frame if necessary (possibly repeatedly)
        while ( internalTimer > frameEndTimes.get(currentFrameIndex) ) 
        {  
            currentFrameIndex++;  
            recentFrameChange = true;
        }
    }
    
    public Image getCurrentImage()
    {
        if (totalFrameCount == 0)
            return null; // no images have been loaded       
        else
            return frameImages.get(currentFrameIndex);
    }
    
    private Image loadImage(String filename)
    {
        Image image;
        try
        {
            URL fileLocation = this.getClass().getClassLoader().getResource(filename);
            image = ImageIO.read( fileLocation );
        } 
        catch (Exception e)
        {
            image = null;
            System.out.println("Could not load file: " + filename);
        }
        return image;
    } 
}
