import javax.swing.JFrame;
import java.awt.*;
import java.awt.image.BufferedImage;
import java.awt.event.KeyListener;
import java.awt.event.KeyEvent;
// useful data structures
import java.util.ArrayList;
import java.util.HashMap;

/* With the Singleton design pattern you can:
 * Ensure that only one instance of a class is created and
 * provide a global point of access to the object
 */

/** provide the stage... the JFrame **/
public class GameManager extends JFrame implements KeyListener
{
    /***************************
     * game specific variables
     ***************************/
    private static int score = 0;
    private static int hiScore = 10000;
    private static int lives = 9;
     
    private static GameManager instance;
    private Scene currentScene;

    private int windowWidth  = 800;
    private int windowHeight = 600;
    
    private final int FRAMES_PER_SECOND = 60;

    // calculate sleep time required to achieve frame rate
    private int sleepTimePerLoop  = 1000 / FRAMES_PER_SECOND; 
    // calculate time change since last update
    private long previousTime = 0; // milliseconds
    private long currentTime  = 0; // milliseconds
    private double dt;             // seconds
    // calculate run time for update and render methods
    private long tempTimeA, tempTimeB, tempTimeC;
    private boolean displayTimeInfo = false;

    // for double buffering
    private BufferedImage backBuffer;
    private Graphics2D g2d;
    // continues the game loop   
    private boolean gameInProgress = true;

    private ArrayList<Integer> activeKeys;
    private HashMap<String,Object> globalVars;

    /* load starting scene for game */
    public static void main(String[] args)
    {   
        // GameManager.get().changeScene( new GameState1() );  
        GameManager.get().changeScene( "GameState1" );  
    }

    public static void main(String sceneName)
    {   
        // GameManager.get().changeScene( new GameState1() );  
        GameManager.get().changeScene( sceneName );  
    }
    ////////////////////
    // singleton code //
    ////////////////////

    private GameManager()
    {    

        setLocation( 0, 0 );
        setSize( windowWidth, windowHeight );
        setTitle( "Boomerang Bill" );
        setResizable( false );
        setFocusable( true );
        setVisible( true );
        setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

        // setup double buffering
        backBuffer = new BufferedImage(windowWidth, windowHeight, BufferedImage.TYPE_INT_RGB);
        g2d = (Graphics2D)backBuffer.createGraphics();

        // set high-quality graphics
        g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING, 
            RenderingHints.VALUE_ANTIALIAS_ON);
        g2d.setRenderingHint(RenderingHints.KEY_INTERPOLATION, 
            RenderingHints.VALUE_INTERPOLATION_BILINEAR);
        g2d.setRenderingHint( RenderingHints.KEY_TEXT_ANTIALIASING,
            RenderingHints.VALUE_TEXT_ANTIALIAS_LCD_HRGB );
        g2d.setRenderingHint(RenderingHints.KEY_RENDERING, 
            RenderingHints.VALUE_RENDER_QUALITY);
            
        // enables this JFrame to process keyboard input
        addKeyListener(this);

        activeKeys = new ArrayList<Integer>();
        globalVars = new HashMap<String,Object>();

        // give focus to JFrame
        try 
        { 
            EventQueue.invokeAndWait( new Runnable() 
                {  public void run() {requestFocusInWindow();}  }  );
        } 
        catch(Exception e) {  }    

        previousTime = System.currentTimeMillis();
        currentTime  = System.currentTimeMillis();
    }

    public static GameManager get()
    {
        if (instance == null) 
            instance = new GameManager();
        return instance;
    }

    /** change active scene, given scene; this method may be called by scenes themselves */
    public static void changeScene(Scene newScene)
    {
        if (instance.currentScene != null) 
            instance.currentScene.unload();
        instance.currentScene = newScene;
        instance.currentScene.load();
        instance.gameLoop();
    }

    /** change active scene, given name of class */
    public static void changeScene(String className)
    {
        try 
        {
            Class theClass  = Class.forName(className);
            Scene newScene = (Scene)theClass.newInstance();
            changeScene( newScene );
            
        }
        catch ( Exception ex )
        {
            System.err.println( "Error loading Scene via name..." );
        }
    }

    /** pass global variables between scenes */
    public static boolean globalVarExists(String varName)
    {   return get().globalVars.containsKey(varName);   }

    public static Object getGlobalVar(String varName)
    {   return get().globalVars.get(varName);   }

    public static Object setGlobalVar(String varName, Object varValue)
    {   return get().globalVars.put(varName, varValue);   }

    ///////////////////////
    // game loop methods //
    ///////////////////////

    public void gameLoop()
    {
        while (gameInProgress)
        {
            try   {  Thread.sleep(sleepTimePerLoop);  }
            catch (InterruptedException e) { }

            previousTime = currentTime;
            currentTime = System.currentTimeMillis();
            dt = (currentTime - previousTime) / 1000.0; 

            tempTimeA = System.currentTimeMillis();
            update(dt);
            tempTimeB = System.currentTimeMillis();
            render();
            tempTimeC = System.currentTimeMillis();

            sleepTimePerLoop = 1000 / FRAMES_PER_SECOND - (int)(tempTimeC - tempTimeA);
            if (sleepTimePerLoop < 0) sleepTimePerLoop = 0;
            if (displayTimeInfo)
                System.out.println(" update: " + (tempTimeB - tempTimeA) + 
                    " render: " + (tempTimeC - tempTimeB) +
                    " new sleep: " + sleepTimePerLoop);
        }
        System.exit(0);
    }

    public void update(double dt) 
    {   
        currentScene.update(dt);
        if ( activeKeys.contains(KeyEvent.VK_ESCAPE) )
            gameInProgress = false;        
    }

    public void render() 
    {   
        currentScene.draw(g2d);
        repaint();
    }

    public void paint(Graphics g) 
    {
        g.drawImage(backBuffer, 0, 0, null);
    }
    ///////////////////
    // input methods //
    ///////////////////

    public void keyPressed(KeyEvent e)
    {
        int i = e.getKeyCode();        
        if (!activeKeys.contains(i)) // do not add duplicates
            activeKeys.add(new Integer(i));   
    }

    public void keyReleased(KeyEvent e)
    {
        int i = e.getKeyCode();
        activeKeys.remove(new Integer(i));
    }

    public void keyTyped(KeyEvent e) 
    {    }

    public static ArrayList getActiveKeys()
    {   return get().activeKeys;   }

    // deactivate input -- good for KeyPressed-like triggering
    public void deactivateKey(int keyCode)
    {
        activeKeys.remove(new Integer(keyCode));
    }
    
    /////game specific methods
     public static int updateScore(int points)
    {
        score += points;
        return score;
    }
    
    public static void setScore(int newScore)
    {
        score = newScore;
    }
    
    public static int getScore()
    {
        return score;
    }
    
    public static int getLives()
    {
        return lives;
    }
    
    public static boolean loseALife()
    {
        if(lives - 1 <= 0)
            return true;
        else
        {
            --lives;
            return false;
        }
    }
    
    public static void setLives(int newLives)
    {
        lives = newLives;
    }

    ////////////////////
    // helper methods //
    ////////////////////
    
    public int[] getDrawBoundaries()
    {
        Insets insets = this.getInsets();
        int xMin = insets.left;
        int xMax = this.getWidth() - insets.right;
        int yMin = insets.top;
        int yMax = this.getHeight() - insets.bottom;  
        int[] boundaries = {xMin, xMax, yMin, yMax};
        return boundaries;
    }
}