import java.net.*;
import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;
import javax.imageio.*;
import java.util.*;

/**
 *   Yet Another Sprite Class...
 *   self-updates and renders.
 *   @author Lee Stemkoski
 */
public class Sprite
{
    private String name = "default"; // used to ID
    
    private double positionX = 0; // pixels
    private double positionY = 0;
    
    private double velocityX = 0; // pixels per second
    private double velocityY = 0;
    
    private double angleFacing = 0; // degrees
    
    // image variables
    private Image  spriteImage;
    public static  BufferedImage blankImage = new BufferedImage(10,10,BufferedImage.TYPE_INT_ARGB);
    private int    imageWidth   = 0;
    private int    imageHeight  = 0;
    
    // if sprite does not move or animate,
    //   set to true to skip updates
    private boolean isStatic = false;
    
    // table for storing images; access by name (string)
    private HashMap<String,Image> imageTable;
    
    // store image transformation data (position, angle, scale) for rendering
    private AffineTransform spriteAT;
    
    // maximum speed of sprite
    private double maxSpeed = 1000;

    // damping: decrease velocity by percentage per second
    private boolean dampingActive     = false;
    private double  dampingPercentage = 0;
    
    // deceleration: decrease velocity by constant per second;
    //   stop when velocity reaches zero. (no negative acceleration)
    private boolean  decelerationActive = false;
    private double   decelerationAmount = 0;
    
    // window boundaries (window width,height minus corresponding insets)
    private int xMin, xMax, yMin, yMax;
    private boolean boundToWindowActive = false;
    private boolean  wrapToWindowActive = false;
    
    // shape used to calculate boundaries for collision detection
    private Rectangle spriteOriginalBox;   // created when image first loaded
    private Shape     spriteBoundingShape; // affine transform of original box
    private Rectangle spriteBoundingBox;   // bounding box of bounding shape (axes horiz&vert)
    private boolean   pixelCollisionActive = true; // use pixel-perfect collision?
    private boolean   collisionDebug = false;
    
    // animation things
    private boolean animationActive = false;
    private Animation spriteAnimation;
    private HashMap<String,Animation> animationTable;

    private boolean destroy = false;
    private boolean animationDestroyActive = false;
    private boolean boundaryDestroyActive = false; 

    /** Initialize sprite with standard basic data. */
    public Sprite(int x, int y, String imageStorageName, Image image )
    {
        positionX = x;
        positionY = y;
        // initialize non-elementary data types
        imageTable = new HashMap<String,Image>();
        animationTable = new HashMap<String,Animation>();
        spriteAT = new AffineTransform();
     
        storeImage( imageStorageName, image );
        loadImageFromStorage( imageStorageName );   
    }
    
    // alternative constructors
    public Sprite(int x, int y, Image image)
    {   this(x, y, "default", image );   }
    public Sprite(int x, int y)
    {   this(x, y, Sprite.blankImage);  }    
    public Sprite()
    {   this(0, 0);  }    
    
    /** Initialize new sprite that clones another sprite. 
     *  @param s    the sprite to be cloned  
     */
    public Sprite(Sprite s)
    {
        name                   = s.name;
        positionX              = s.positionX;
        positionY              = s.positionY;
        velocityX              = s.velocityX;
        velocityY              = s.velocityY;
        angleFacing            = s.angleFacing;
        spriteImage            = s.spriteImage;
        imageWidth             = s.imageWidth;
        imageHeight            = s.imageHeight;
        imageTable             = new HashMap<String,Image>(s.imageTable);
        spriteAT               = new AffineTransform(s.spriteAT);
        isStatic               = s.isStatic;
        maxSpeed               = s.maxSpeed;
        dampingActive          = s.dampingActive;
        dampingPercentage      = s.dampingPercentage;
        decelerationActive     = s.decelerationActive;
        decelerationAmount     = s.decelerationAmount;    
        xMin                   = s.xMin; 
        xMax                   = s.xMax;
        yMin                   = s.yMin;
        yMax                   = s.yMax;
        boundToWindowActive    = s.boundToWindowActive;
        wrapToWindowActive     = s.wrapToWindowActive;
        spriteOriginalBox      = new Rectangle( s.spriteOriginalBox ); 
        spriteBoundingShape    = new Path2D.Double( s.spriteBoundingShape ); 
        spriteBoundingBox      = new Rectangle( s.spriteBoundingBox ); 
        animationActive        = s.animationActive;
        spriteAnimation        = new Animation( s.spriteAnimation );
        animationTable         = new HashMap<String,Animation>(s.animationTable);
        animationDestroyActive = s.animationDestroyActive;
        boundaryDestroyActive  = s.boundaryDestroyActive;
    }
    
    public void setName(String s)
    {   name = s;   }
    
    public String getName()
    {   return name;   }
    
    ///////////////////////////
    // image-related methods //
    ///////////////////////////
    
    
    public void storeImage(String storageName, Image storageImage)
    {   
        imageTable.put( storageName, storageImage );   
        loadImageFromStorage( storageName );
    }
    
    public void storeImage(Image storeImage)
    {   storeImage("default", storeImage);   }
    
    public void storeImage(String storageName, String imageFileName)
    {   storeImage( storageName, loadImage(imageFileName) );   }
    
    public void loadImageFromStorage(String storageName)
    {   setCurrentImage( imageTable.get(storageName) );   }
    
    public void setCurrentImage(Image newImage)
    {
        spriteImage = newImage;
        // image-related calculations
        imageWidth   = spriteImage.getWidth(null);
        imageHeight  = spriteImage.getHeight(null);
        // for collision detection
        spriteOriginalBox   = new Rectangle(0, 0, imageWidth, imageHeight);
        spriteBoundingShape = new Rectangle(0, 0, imageWidth, imageHeight);
        spriteBoundingBox   = new Rectangle(0, 0, imageWidth, imageHeight);
    }
    
    public int getWidth()
    {   return imageWidth;   }
    
    public int getHeight()
    {   return imageHeight;   }
   
    ///////////////////////////////
    // animation-related methods //
    ///////////////////////////////
    
    public void setAnimationActive(boolean b)
    {   animationActive = b;   }
    
    public void storeAnimation(String storageName, Animation storageAnimation)
    {   
        animationTable.put( storageName, storageAnimation );  
        loadAnimationFromStorage( storageName );
    }
    
    public void storeAnimation(Animation storageAnimation)
    {   
        animationTable.put( "default", storageAnimation );  
        loadAnimationFromStorage( "default" );
    }
    
    public void loadAnimationFromStorage(String animationName)
    {   
        spriteAnimation = animationTable.get( animationName ); 
        loadImageFromAnimation();
        setAnimationActive(true);
    }
        
    public void loadImageFromAnimation()
    {   setCurrentImage( spriteAnimation.getCurrentImage() );   }
    
    //////////////////////////////
    // position-related methods //
    //////////////////////////////
    
    public void setPosition(double x, double y)
    {
        positionX = x;
        positionY = y;
    }
    public void setPositionX(double x)
    {   positionX = x;   }
    public void setPositionY(double y)
    {   positionY = y;   }
    public void setAngleFacing(double a)
    {   angleFacing = a;   }
    
    public double getPositionX()
    {   return positionX;   }
    public double getPositionY()
    {   return positionY;   }
    public double getAngleFacing()
    {   return angleFacing;   }
    public double getImageCenterX()
    {   return positionX + imageWidth/2.0;   }
    public double getImageCenterY()
    {   return positionY + imageHeight/2.0;   }
    
    public void addPositionX(double x)
    {   positionX += x;   }
    public void addPositionY(double y)
    {   positionY += y;   }
    public void addAngleFacing(double a)
    {   angleFacing += a;   }
  
    public void alignToCenter(Sprite otherSprite)
    {
        this.positionX = otherSprite.getImageCenterX() - this.getWidth()/2;
        this.positionY = otherSprite.getImageCenterY() - this.getHeight()/2;
    }
    //////////////////////////////
    // velocity-related methods //
    //////////////////////////////
    
    /** Set sprite velocity in terms of speed and (global) angle.
     *  
     */
    public void setVelocity(double speed, double angle)
    {
        velocityX = speed * Math.cos(Math.toRadians(angle));
        velocityY = speed * Math.sin(Math.toRadians(angle));
    }
    
    public void setVelocityLocal(double speed, double angle)
    {   setVelocity(speed, angle + angleFacing);   }
    
    /** Change sprite speed but maintain current angle of motion. */
    public void setVelocitySpeed(double speed)
    {   setVelocity(speed, getVelocityAngle());   }
    
    /** Change sprite angle of motion but maintain current speed. */
    public void setVelocityAngle(double angle)
    {   setVelocity(getVelocitySpeed(), angle);   }
    
    /** Change sprite angle of motion (relative to image angle) but maintain current speed. */
    public void setVelocityAngleLocal(double angle)
    {   setVelocity(getVelocitySpeed(), angle + angleFacing);   }
    
    public double getVelocityX()
    { return this.velocityX; }
    
    public double getVelocityY()
    { return this.velocityY; }
        
   
    public double getVelocitySpeed()
    {   return Math.sqrt(velocityX*velocityX + velocityY*velocityY);   }
    
    public double getVelocityAngle()
    {   return Math.toDegrees(Math.atan2(velocityY, velocityX));   }

    public void addVelocity(double speed, double angle)
    {
        velocityX += speed * Math.cos(Math.toRadians(angle));
        velocityY += speed * Math.sin(Math.toRadians(angle));
    }
    
    public void addVelocityLocal(double speed, double angle)
    {   addVelocity(speed, angle + angleFacing);   }
    
    public void setDampingActive(boolean b)
    {   dampingActive = b;   }
    
    public void setDampingPercentage(double p)
    {   dampingPercentage = p;   }
    
    public void setDecelerationActive(boolean b)
    {   decelerationActive = b;   }
    
    public void setDecelerationAmount(double s)
    {   decelerationAmount = s;   }
    
    public void setMaxSpeed(double s)
    {   maxSpeed = s;   }
    
    public void setVelocityX(double x)
    {
        this.velocityX = x;
    }
    public void setVelocityY(double newY)
    {
        this.velocityY = newY;
    }
    
    public void flipVelocityX()
    {
        this.velocityX *= -1;
    }
    
    public void flipVelocityY()
    {
        this.velocityY *= -1;
    }

    ////////////////////////////
    // window-related methods //
    ////////////////////////////
    
    public void setWindowBounds(int xMin, int xMax, int yMin, int yMax)
    {
        this.xMin = xMin;
        this.xMax = xMax;
        this.yMin = yMin;
        this.yMax = yMax;
    }
    
    public void setWindowBounds(int[] bounds)
    {   setWindowBounds(bounds[0], bounds[1], bounds[2], bounds[3]);   }
    
    public void setBoundToWindow(boolean b)
    {   boundToWindowActive = b;   }
    
    private void boundToWindow()
    {
        if (positionX < xMin) 
        {
            positionX = xMin;
            velocityX = 0;
        }
        if (positionX + imageWidth > xMax) 
        {
            positionX = xMax - imageWidth;
            velocityX = 0;
        }
        if (positionY < yMin) 
        {
            positionY = yMin;
            velocityY = 0;
        }
        if (positionY + imageHeight > yMax) 
        {
            positionY = yMax - imageHeight;
            velocityY = 0;
        }
    }
    
    public void setWrapToWindow(boolean b)
    {   wrapToWindowActive = b;   }
    
    private void wrapToWindow()
    {
        if (positionX + imageWidth < xMin) 
            positionX = xMax;
        if (positionX > xMax) 
            positionX = xMin - imageWidth;
        if (positionY + imageHeight < yMin) 
            positionY = yMax;
        if (positionY > yMax) 
            positionY = yMin - imageHeight;
    }
    
    
    
    public boolean isOutsideWindow()
    {
        return (positionX + imageWidth  < xMin) || (positionX > xMax) 
            || (positionY + imageHeight < yMin) || (positionY > yMax);
    }
    
    ////////////
    // update //
    ////////////
    
    public void update()
    {   update(0);   }
        
    public void setStatic(boolean b)
    {   isStatic = b;   }
    
    public void update(double dt)
    {
        if (destroy) return;
        
        if (isStatic)
            return;
            
        if (animationActive)
        {
            spriteAnimation.update( (int)(1000*dt) );
            
            if ( spriteAnimation.frameChanged() )
                loadImageFromAnimation();
                
            if ( animationDestroyActive && spriteAnimation.isAnimationFinished() )
                destroy = true;
        }
        
        // apply deceleration/gravity...
        // apply damping -- percent of velocity to lose each second.
        // http://www.digitalrune.com/Support/Blog/tabid/719/EntryId/78/Damping-in-Computer-Games.aspx
        // http://gamedev.stackexchange.com/questions/4673/not-sure-how-to-handle-deceleration
        if (dampingActive)
        {
            double dampingFactor = Math.exp(-dampingPercentage * dt);
            velocityX *= dampingFactor;
            velocityY *= dampingFactor;
        }
        
        if (decelerationActive)
        {
            double speed = getVelocitySpeed();
            if (decelerationAmount*dt > speed)
                setVelocitySpeed(0);
            else
                setVelocitySpeed(speed - decelerationAmount*dt);
        }
        
        if ( getVelocitySpeed() > maxSpeed )
            setVelocity( maxSpeed, getVelocityAngle() );
        
        // update position
        positionX += velocityX * dt;
        positionY += velocityY * dt;
        
        if (boundToWindowActive)
            boundToWindow();
            
        if (wrapToWindowActive)
            wrapToWindow();
        
        if (boundaryDestroyActive && isOutsideWindow() )
            destroy = true;
            
        // update transform
        spriteAT.setToIdentity();
        spriteAT.translate(positionX, positionY);
        spriteAT.rotate(Math.toRadians(angleFacing), imageWidth/2, imageHeight/2); 
        // update bounding regions
        spriteBoundingShape = spriteAT.createTransformedShape(spriteOriginalBox);
        spriteBoundingBox   = spriteBoundingShape.getBounds();
    }
    
    ////////////
    // render //
    ////////////
    
    public void render(Graphics2D g2d)
    {   
        if (destroy) return;
        
        g2d.drawImage( spriteImage, spriteAT, null);     
        if (collisionDebug)
        {
            g2d.setColor( Color.red );
            g2d.draw( spriteBoundingShape );
            g2d.setColor( Color.yellow );
            g2d.draw( spriteBoundingBox );
            g2d.setColor( Color.green );
            g2d.fillRect( (int)(positionX - 2), (int)(positionY - 2), 5, 5 );
            g2d.fillRect( (int)getImageCenterX() - 2, (int)getImageCenterY() - 2, 5, 5 );
        }
    }

    /////////////////////
    // removal methods //
    /////////////////////
    
    public void setDestroy(boolean b)
    {   destroy = b;   }
    
    public void setAnimationDestroyActive(boolean b)
    {   animationDestroyActive = b;   }
    
    public void setBoundaryDestroyActive(boolean b)
    {   boundaryDestroyActive = b;   }
    
    /** Remove sprites with (destroy = true) from list. */
    public static void removeDestroyedSprites(ArrayList<Sprite> spriteList)
    {
        // make a copy of the list as an array (to avoid concurrent modification exception)    
        ArrayList<Integer> toRemove = new ArrayList<Integer>();
        for (Sprite s : spriteList)
        {   
            if (s.destroy)
                toRemove.add( spriteList.indexOf(s) );
        } 
        try
        {
            for (int i: toRemove)
                spriteList.remove(i);
        }
        catch (Exception e) {  }  // out of bounds
    }
    
    public static void removeDestroyedSprites2(ArrayList<Object> masterList)
    {
        // make a copy of the list as an array (to avoid concurrent modification exception)    
        ArrayList<Integer> toRemove = new ArrayList<Integer>();
        for (Object o : masterList)
        {   
            if (o instanceof Sprite)
            {   
                Sprite s = (Sprite)o;
                if (s.destroy)
                    toRemove.add( masterList.indexOf(s) );
            }
            
            if (o instanceof ArrayList)
            {
                ArrayList<Sprite> al = (ArrayList<Sprite>)o;
                removeDestroyedSprites(al);
            }
        } 
        try
        {
            for (int i: toRemove)
                masterList.remove(i);
        }
        catch (Exception e) {  }  // out of bounds
    }
    // helper method
    public Image loadImage(String filename)
    {
        BufferedImage image;
        try
        {
            URL fileLocation = this.getClass().getClassLoader().getResource(filename);
            image = ImageIO.read( fileLocation );
        } 
        catch (Exception e)
        {
            image = null;
            System.out.println("Could not load file: " + filename);
        }
        return image;
    } 
    
    /////////////////////////
    // collision detection //
    /////////////////////////
    
    public Shape getBoundingShape()
    {
        if (spriteBoundingShape == null)  return null;
        else  return spriteBoundingShape;
    }
    
    /**
     *  @return a rectangle that contains the transformed image;
     *  specifically, the rectangle that contains the BoundingShape
     *  @see #getBoundingShape()
     *  
     */
    public Rectangle getBoundingBox()
    {
        if (spriteBoundingBox == null)  return null;
        else  return spriteBoundingBox;
    }
    
    public void setPixelCollisionActive(boolean b)
    {   pixelCollisionActive = b;   }
    
    public boolean intersects(Sprite otherSprite)
    {
        boolean isBoxCollision = spriteBoundingShape.intersects( otherSprite.getBoundingBox() );
        
        if ( !isBoxCollision )
            return false;
        if ( !pixelCollisionActive )
            return true;
        else
            return pixelPerfectCollision(otherSprite);
    }
    
    public void moveOff(Sprite otherSprite)
    {
        //double awayVectorX = this.positionX - otherSprite.positionX;
        //double awayVectorY = this.positionY - otherSprite.positionY;
        double awayVectorX = getImageCenterX() - otherSprite.getImageCenterX(); 
        double awayVectorY = getImageCenterY() - otherSprite.getImageCenterY(); 
        // normalize to 1 pixel length
        double vecLength = Math.sqrt(awayVectorX*awayVectorX + awayVectorY*awayVectorY);
        awayVectorX /= vecLength;
        awayVectorY /= vecLength;
        int moveAttempt = 0;
        while ( this.intersects(otherSprite) )
        {
            this.positionX += 0.5 * awayVectorX;
            this.positionY += 0.5 * awayVectorY;
            // update transform
            spriteAT.setToIdentity();
            spriteAT.translate(positionX, positionY);
            spriteAT.rotate(Math.toRadians(angleFacing), imageWidth/2, imageHeight/2); 
            // update bounding regions
            spriteBoundingShape = spriteAT.createTransformedShape(spriteOriginalBox);
            spriteBoundingBox   = spriteBoundingShape.getBounds();
            moveAttempt++;
            if (moveAttempt > 100)
            {
                // System.out.println(this.getName() + " is stuck within " + otherSprite.getName());
                break;
            }
        }
    }
    
    // ONLY WORKS WITH ROUGHLY CIRCULAR CONVEX OBJECTS. 
    //      TRIANGLES ARE ESPECIALLY BAD.
    public void bounceOff(Sprite otherSprite)
    {
        double tempX = this.getImageCenterX() - otherSprite.getImageCenterX();
        double tempY = this.getImageCenterY() - otherSprite.getImageCenterY();
        // angle that the tangent line to the circle makes with the x-axis
        double alpha = Math.atan2( tempX, -tempY ); // radians
        // angle of reflection = angle of refraction
        this.setVelocityAngle( Math.toDegrees(2 * alpha) - this.getVelocityAngle() );
    }
    
    public Image getCurrentImageRotated() // used for pixel-perfect collision detection
    {
        BufferedImage tempBuffer = new BufferedImage(
            spriteBoundingBox.width, spriteBoundingBox.height,
            BufferedImage.TYPE_INT_ARGB);
            
        Graphics2D tempGraphics = tempBuffer.createGraphics();
        tempGraphics.rotate( Math.toRadians(angleFacing), imageWidth/2, imageHeight/2);
        tempGraphics.drawImage(spriteImage,0,0,null);
        return tempBuffer;
    }

     /*
     *  pixelPerfectCollision(); first determines the area where the sprites collides
     *  AKA the collision-rectangle. It then grabs the pixels from both sprites
     *  which are inside the rectangle. It then checks every pixel from the arrays
     *  given by grabPixels();, and if 2 pixels at the same position are opaque,
     *  (alpha value over 0) it will return true. Otherwise it will return false.
     */
    public boolean pixelPerfectCollision(Sprite otherSprite) 
    {
        Rectangle r1 = this.getBoundingBox();
        Rectangle r2 = otherSprite.getBoundingBox();
        int cornerTopX=-1;
        int cornerTopY=-1;
        int cornerBottomX = 1;
        int cornerBottomY = 1;

        cornerTopX = (r1.x > r2.x)
            ? r1.x : r2.x;
        cornerBottomX = ((r1.x+r1.width) < (r2.x+r2.width)) 
            ? (r1.x + r1.width) : (r2.x + r2.width);
        cornerTopY = (r1.y>r2.y)
            ? r1.y : r2.y;
        cornerBottomY = ((r1.y+r1.height) < (r2.y+r2.height))
            ? (r1.y+r1.height) : (r2.y+r2.height);
 
        //Determine the width and height of the collision rectangle
        int width=cornerBottomX-cornerTopX;
        int height=cornerBottomY-cornerTopY;
        
        //Create arrays to hold the pixels
        int[] pixelArray1 = new int[width*height];
        int[] pixelArray2 = new int[width*height];
 
        // PixelGrabber: grabs a rectangular subset of pixels from an image
        //    and stores them in an array.
        // parameters: upperLeftX,upperLeftY,subsetWidth,subsetHeight,
        //    storageArray,firstPixelArrayStorageOffset,distanceBetweenRowsInArray
        PixelGrabber pg1 = new PixelGrabber(this.getCurrentImageRotated(), 
            cornerTopX-(int)getPositionX(), cornerTopY-(int)getPositionY(), 
            width, height, pixelArray1, 0, width);
        PixelGrabber pg2 = new PixelGrabber(otherSprite.getCurrentImageRotated(), 
            cornerTopX-(int)otherSprite.getPositionX(), cornerTopY-(int)otherSprite.getPositionY(), 
            width, height, pixelArray2, 0, width);
 
        //Grab the pixels
        try 
        {
            pg1.grabPixels();
            pg2.grabPixels();
        } 
        catch (InterruptedException ex) {  }
 
        //Check if pixels at the same spot from both arrays are not transparent.
        for (int i = 0; i < pixelArray1.length; i++) 
        {
            int a1 = (pixelArray1[i] >>> 24) & 0xff;
            int a2 = (pixelArray2[i] >>> 24) & 0xff;
            if (a1 > 50 && a2 > 50)  // both pixels are not transparent
                return true;
        }
       
        return false;
    }
}